#!/usr/bin/env python

"""
Interpretation and preparation of iMIP content, together with a content handling
mechanism employed by specific recipients.
"""

from datetime import date, datetime, timedelta
from email.mime.text import MIMEText
from imiptools.period import have_conflict, insert_period, remove_period
from pytz import timezone, UnknownTimeZoneError
from vCalendar import parse, ParseError, to_dict
from vRecurrence import get_parameters, get_rule
import imip_store
import re

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# iCalendar date and datetime parsing (from DateSupport in MoinSupport).

date_icalendar_regexp_str = ur'(?P<year>[0-9]{4})(?P<month>[0-9]{2})(?P<day>[0-9]{2})'
datetime_icalendar_regexp_str = date_icalendar_regexp_str + \
    ur'(?:' \
    ur'T(?P<hour>[0-2][0-9])(?P<minute>[0-5][0-9])(?P<second>[0-6][0-9])' \
    ur'(?P<utc>Z)?' \
    ur')?'

match_date_icalendar = re.compile(date_icalendar_regexp_str, re.UNICODE).match
match_datetime_icalendar = re.compile(datetime_icalendar_regexp_str, re.UNICODE).match

# Content interpretation.

def get_items(d, name, all=True):

    """
    Get all items from 'd' with the given 'name', returning single items if
    'all' is specified and set to a false value and if only one value is
    present for the name. Return None if no items are found for the name.
    """

    if d.has_key(name):
        values = d[name]
        if not all and len(values) == 1:
            return values[0]
        else:
            return values
    else:
        return None

def get_item(d, name):
    return get_items(d, name, False)

def get_value_map(d, name):

    """
    Return a dictionary for all items in 'd' having the given 'name'. The
    dictionary will map values for the name to any attributes or qualifiers
    that may have been present.
    """

    items = get_items(d, name)
    if items:
        return dict(items)
    else:
        return {}

def get_values(d, name, all=True):
    if d.has_key(name):
        values = d[name]
        if not all and len(values) == 1:
            return values[0][0]
        else:
            return map(lambda x: x[0], values)
    else:
        return None

def get_value(d, name):
    return get_values(d, name, False)

def get_utc_datetime(d, name):
    value, attr = get_item(d, name)
    dt = get_datetime(value, attr)
    return to_utc_datetime(dt)

def to_utc_datetime(dt):
    if not dt:
        return None
    elif isinstance(dt, datetime):
        return dt.astimezone(timezone("UTC"))
    else:
        return dt

def to_timezone(dt, name):
    try:
        tz = name and timezone(name) or None
    except UnknownTimeZoneError:
        tz = None
    if tz is not None:
        if not dt.tzinfo:
            return tz.localize(dt)
        else:
            return dt.astimezone(tz)
    else:
        return dt

def format_datetime(dt):
    if not dt:
        return None
    elif isinstance(dt, datetime):
        if dt.tzname() == "UTC":
            return dt.strftime("%Y%m%dT%H%M%SZ")
        else:
            return dt.strftime("%Y%m%dT%H%M%S")
    else:
        return dt.strftime("%Y%m%d")

def get_address(value):
    return value.startswith("mailto:") and value[7:] or value

def get_uri(value):
    return value.startswith("mailto:") and value or "mailto:%s" % value

def get_datetime(value, attr=None):

    """
    Return a datetime object from the given 'value' in iCalendar format, using
    the 'attr' mapping (if specified) to control the conversion.
    """

    if not attr or attr.get("VALUE") in (None, "DATE-TIME"):
        m = match_datetime_icalendar(value)
        if m:
            dt = datetime(
                int(m.group("year")), int(m.group("month")), int(m.group("day")),
                int(m.group("hour")), int(m.group("minute")), int(m.group("second"))
                )

            # Impose the indicated timezone.
            # NOTE: This needs an ambiguity policy for DST changes.

            return to_timezone(dt, m.group("utc") and "UTC" or attr and attr.get("TZID") or None)

    if not attr or attr.get("VALUE") == "DATE":
        m = match_date_icalendar(value)
        if m:
            return date(
                int(m.group("year")), int(m.group("month")), int(m.group("day"))
                )
    return None

# NOTE: Need to expose the 100 day window for recurring events in the
# NOTE: configuration.

def get_periods(obj, window_size=100):

    """
    Return periods for the given object 'obj', confining materialised periods
    to the given 'window_size' in days starting from the present moment.
    """

    dtstart = get_utc_datetime(obj, "DTSTART")
    dtend = get_utc_datetime(obj, "DTEND")

    # NOTE: Need also DURATION support.

    duration = dtend - dtstart

    # Recurrence rules create multiple instances to be checked.
    # Conflicts may only be assessed within a period defined by policy
    # for the agent, with instances outside that period being considered
    # unchecked.

    window_end = datetime.now() + timedelta(window_size)

    # NOTE: Need also RDATE and EXDATE support.

    rrule = get_value(obj, "RRULE")

    if rrule:
        selector = get_rule(dtstart, rrule)
        parameters = get_parameters(rrule)
        periods = []
        for start in selector.materialise(dtstart, window_end, parameters.get("COUNT"), parameters.get("BYSETPOS")):
            start = datetime(*start, tzinfo=timezone("UTC"))
            end = start + duration
            periods.append((format_datetime(start), format_datetime(end)))
    else:
        periods = [(format_datetime(dtstart), format_datetime(dtend))]

    return periods

def update_freebusy(freebusy, attendee, periods, transp, uid, store):

    """
    For the given 'attendee', update the free/busy details with the given
    'periods', 'transp' setting and 'uid' in the 'store'.
    """

    remove_period(freebusy, uid)

    for start, end in periods:
        insert_period(freebusy, (start, end, uid, transp))

    store.set_freebusy(attendee, freebusy)

def can_schedule(freebusy, periods, uid):

    """
    Return whether the 'freebusy' list can accommodate the given 'periods'
    employing the specified 'uid'.
    """

    for conflict in have_conflict(freebusy, periods, True):
        start, end, found_uid, found_transp = conflict
        if found_uid != uid:
            return False

    return True

# Handler mechanism objects.

def handle_itip_part(part, senders, recipients, handlers, messenger):

    """
    Handle the given iTIP 'part' from the given 'senders' for the given
    'recipients' using the given 'handlers' and information provided by the
    given 'messenger'. Return a list of responses, each response being a tuple
    of the form (is-outgoing, message-part).
    """

    method = part.get_param("method")

    # Decode the data and parse it.

    f = StringIO(part.get_payload(decode=True))

    itip = parse_object(f, part.get_content_charset(), "VCALENDAR")

    # Ignore the part if not a calendar object.

    if not itip:
        return []

    # Require consistency between declared and employed methods.

    if get_value(itip, "METHOD") == method:

        # Look for different kinds of sections.

        all_results = []

        for name, cls in handlers:
            for details in get_values(itip, name) or []:

                # Dispatch to a handler and obtain any response.

                handler = cls(details, senders, recipients, messenger)
                result = methods[method](handler)()

                # Aggregate responses for a single message.

                if result:
                    response_method, part = result
                    outgoing = method != response_method
                    all_results.append((outgoing, part))

        return all_results

    return []

def parse_object(f, encoding, objtype=None):

    """
    Parse the iTIP content from 'f' having the given 'encoding'. If 'objtype' is
    given, only objects of that type will be returned. Otherwise, the root of
    the content will be returned as a dictionary with a single key indicating
    the object type.

    Return None if the content was not readable or suitable.
    """

    try:
        try:
            doctype, attrs, elements = obj = parse(f, encoding=encoding)
            if objtype and doctype == objtype:
                return to_dict(obj)[objtype][0]
            elif not objtype:
                return to_dict(obj)
        finally:
            f.close()

    # NOTE: Handle parse errors properly.

    except (ParseError, ValueError):
        pass

    return None

def to_part(method, calendar):

    """
    Write using the given 'method', the 'calendar' details to a MIME
    text/calendar part.
    """

    encoding = "utf-8"
    out = StringIO()
    try:
        imip_store.to_stream(out, imip_store.make_calendar(calendar, method), encoding)
        part = MIMEText(out.getvalue(), "calendar", encoding)
        part.set_param("method", method)
        return part

    finally:
        out.close()

class Handler:

    "General handler support."

    def __init__(self, details, senders=None, recipients=None, messenger=None):

        """
        Initialise the handler with the 'details' of a calendar object and the
        'senders' and 'recipients' of the object (if specifically indicated).
        """

        self.details = details
        self.senders = senders and set(senders)
        self.recipients = recipients and set(recipients)
        self.messenger = messenger

        self.uid = get_value(details, "UID")
        self.sequence = get_value(details, "SEQUENCE")
        self.dtstamp = get_value(details, "DTSTAMP")

        self.store = imip_store.FileStore()

        try:
            self.publisher = imip_store.FilePublisher()
        except OSError:
            self.publisher = None

    # Access to calendar structures and other data.

    def get_items(self, name, all=True):
        return get_items(self.details, name, all)

    def get_item(self, name):
        return get_item(self.details, name)

    def get_value_map(self, name):
        return get_value_map(self.details, name)

    def get_values(self, name, all=True):
        return get_values(self.details, name, all)

    def get_value(self, name):
        return get_value(self.details, name)

    def get_utc_datetime(self, name):
        return get_utc_datetime(self.details, name)

    def get_periods(self):
        return get_periods(self.details)

    def update_freebusy(self, freebusy, attendee, periods):
        return update_freebusy(freebusy, attendee, periods, self.get_value("TRANSP"), self.uid, self.store)

    def can_schedule(self, freebusy, periods):
        return can_schedule(freebusy, periods, self.uid)

    def filter_by_senders(self, values):
        addresses = map(get_address, values)
        if self.senders:
            return self.senders.intersection(addresses)
        else:
            return addresses

    def filter_by_recipients(self, values):
        addresses = map(get_address, values)
        if self.recipients:
            return self.recipients.intersection(addresses)
        else:
            return addresses

    def require_organiser_and_attendees(self, from_organiser=True):

        """
        Return the organiser and attendees for the current object, filtered by
        the recipients of interest. Return None if no identities are eligible.
        """

        attendee_map = self.get_value_map("ATTENDEE")
        organiser = self.get_item("ORGANIZER")

        # Only provide details for recipients who are also attendees.

        filter_fn = from_organiser and self.filter_by_recipients or self.filter_by_senders

        attendees = {}
        for attendee in map(get_uri, filter_fn(attendee_map)):
            attendees[attendee] = attendee_map[attendee]

        if not attendees or not organiser:
            return None

        return organiser, attendees

    def validate_identities(self, items):

        """
        Validate the 'items' against the known senders, obtaining sent-by
        addresses from attributes provided by the items.
        """

        # Reject organisers that do not match any senders.

        identities = []

        for value, attr in items:
            identities.append(value)
            sent_by = attr.get("SENT-BY")
            if sent_by:
                identities.append(sent_by)

        return self.filter_by_senders(identities)

    def get_object(self, user, objtype):

        """
        Return the stored object to which the current object refers for the
        given 'user' and for the given 'objtype'.
        """

        f = self.store.get_event(user, self.uid)
        obj = f and parse_object(f, "utf-8", objtype)
        return obj

    def have_new_object(self, attendee, objtype, obj=None):

        """
        Return whether the current object is new to the 'attendee' for the
        given 'objtype'.
        """

        obj = obj or self.get_object(attendee, objtype)

        # If found, compare SEQUENCE and potentially DTSTAMP.

        if obj:
            sequence = get_value(obj, "SEQUENCE")
            dtstamp = get_value(obj, "DTSTAMP")

            # If the request refers to an older version of the object, ignore
            # it.

            old_dtstamp = self.dtstamp < dtstamp

            if sequence is not None and (
                int(self.sequence) < int(sequence) or
                int(self.sequence) == int(sequence) and old_dtstamp
                ) or old_dtstamp:

                return False

        return True

# Handler registry.

methods = {
    "ADD"            : lambda handler: handler.add,
    "CANCEL"         : lambda handler: handler.cancel,
    "COUNTER"        : lambda handler: handler.counter,
    "DECLINECOUNTER" : lambda handler: handler.declinecounter,
    "PUBLISH"        : lambda handler: handler.publish,
    "REFRESH"        : lambda handler: handler.refresh,
    "REPLY"          : lambda handler: handler.reply,
    "REQUEST"        : lambda handler: handler.request,
    }

# vim: tabstop=4 expandtab shiftwidth=4
